@@ -5,6 +5,7 @@
  - Licensed under the GNU GPL version 3 or higher.
  -}
 
+{-# LANGUAGE CPP #-}
 {-# LANGUAGE GeneralizedNewtypeDeriving #-}
 {-# LANGUAGE PackageImports #-}
 
@@ -16,12 +17,13 @@ import qualified Data.Text as T
 import Data.Text.Encoding (encodeUtf8)
 import Data.Either
 import Data.Monoid
-import System.Environment (getArgs)
-import Control.Exception (try, SomeException)
+import Options.Applicative
 import Text.Show.Pretty
 import "mtl" Control.Monad.State.Strict
-import qualified Github.Data.Readable as Github
 import qualified Github.Repos as Github
+#if MIN_VERSION_github(0,9,0)
+import qualified Github.Auth as Github
+#endif
 import qualified Github.Repos.Forks as Github
 import qualified Github.PullRequests as Github
 import qualified Github.Repos.Watching as Github
@@ -216,7 +218,7 @@ forksStore = simpleHelper "forks" Github.forksFor' $ \req fs -> do
 	mapM_ addFork fs
 
 forValues :: (Request -> v -> Backup ()) -> Request -> [v] -> Backup ()
-forValues handle req vs = forM_ vs (handle req)
+forValues a req vs = forM_ vs (a req)
 
 type ApiCall v = Maybe Github.GithubAuth -> String -> String -> IO (Either Github.Error v)
 type ApiWith v b = Maybe Github.GithubAuth -> String -> String -> b -> IO (Either Github.Error v)
@@ -225,24 +227,24 @@ type Handler v = Request -> v -> Backup ()
 type Helper = Request -> Backup ()
 
 simpleHelper :: FilePath -> ApiCall v -> Handler v -> Helper
-simpleHelper dest call handle req@(RequestSimple _ (GithubUserRepo user repo)) =
+simpleHelper dest call handler req@(RequestSimple _ (GithubUserRepo user repo)) =
 	deferOn dest req $ do
 		auth <- getState gitHubAuth
-		either (failedRequest req) (handle req) =<< liftIO (call auth user repo)
+		either (failedRequest req) (handler req) =<< liftIO (call auth user repo)
 simpleHelper _ _ _ r = badRequest r
 
 withHelper :: FilePath -> ApiWith v b -> b -> Handler v -> Helper
-withHelper dest call b handle req@(RequestSimple _ (GithubUserRepo user repo)) =
+withHelper dest call b handler req@(RequestSimple _ (GithubUserRepo user repo)) =
 	deferOn dest req $ do
 		auth <- getState gitHubAuth
-		either (failedRequest req) (handle req) =<< liftIO (call auth user repo b)
+		either (failedRequest req) (handler req) =<< liftIO (call auth user repo b)
 withHelper _ _ _ _ r = badRequest r
 
 numHelper :: FilePath -> ApiNum v -> (Int -> Handler v) -> Helper
-numHelper dest call handle req@(RequestNum _ (GithubUserRepo user repo) num) =
+numHelper dest call handler req@(RequestNum _ (GithubUserRepo user repo) num) =
 	deferOn dest req $ do
 		auth <- getState gitHubAuth
-		either (failedRequest req) (handle num req) =<< liftIO (call auth user repo num)
+		either (failedRequest req) (handler num req) =<< liftIO (call auth user repo num)
 numHelper _ _ _ r = badRequest r
 
 badRequest :: Request -> a
@@ -259,7 +261,7 @@ deferOn f req a = ifM (ingit $ storeLocation f req)
 	ingit f' = do
 		h <- getCatFileHandle
 		liftIO $ isJust <$> catObjectDetails h
-			(Git.Types.Ref $ show branchname ++ ":" ++ f')
+			(Git.Types.Ref $ Git.Types.fromRef branchname ++ ":" ++ f')
 
 getCatFileHandle :: Backup CatFileHandle
 getCatFileHandle = go =<< getState catFileHandle
@@ -313,7 +315,7 @@ gitHubPairs = filter (not . wiki ) . mapMaybe check . Git.Types.remotes
 		| otherwise = Nothing
 	  where
 		rest = drop (length prefix) u
-		bits = split "/" rest
+		bits = filter (not . null) $ split "/" rest
 	dropdotgit s
 		| ".git" `isSuffixOf` s = take (length s - length ".git") s
 		| otherwise = s
@@ -346,14 +348,14 @@ commitWorkDir = do
 				-- Reset index to current content of github
 				-- branch. Does not touch work tree.
 				Git.Command.run
-					[Param "reset", Param "-q", Param $ show branchref, File "." ] r
+					[Param "reset", Param "-q", Param $ Git.Types.fromRef branchref, File "." ] r
 				-- Stage workDir files into the index.
 				h <- hashObjectStart r
 				Git.UpdateIndex.streamUpdateIndex r
 					[genstream dir h]
 				hashObjectStop h
 				-- Commit
-				void $ Git.Branch.commit "github-backup" fullname [branchref] r
+				void $ Git.Branch.commit Git.Branch.AutomaticCommit False "github-backup" fullname [branchref] r
 				removeDirectoryRecursive dir
   where
   	genstream dir h streamer = do
@@ -370,11 +372,11 @@ getBranch = maybe (hasOrigin >>= create) return =<< branchsha
   where
   	create True = do
 		inRepo $ Git.Command.run
-			[Param "branch", Param $ show branchname, Param $ show originname]
-		fromMaybe (error $ "failed to create " ++ show branchname)
+			[Param "branch", Param $ Git.Types.fromRef branchname, Param $ Git.Types.fromRef originname]
+		fromMaybe (error $ "failed to create " ++ Git.Types.fromRef branchname)
 			<$> branchsha
 	create False = withIndex $
-		inRepo $ Git.Branch.commit "branch created" fullname []
+		inRepo $ Git.Branch.commitAlways Git.Branch.AutomaticCommit "branch created" fullname []
 	branchsha = inRepo $ Git.Ref.sha fullname
 
 {- Runs an action with a different index file, used for the github branch. -}
@@ -393,10 +395,10 @@ branchname :: Git.Ref
 branchname = Git.Ref "github"
 
 fullname :: Git.Ref
-fullname = Git.Ref $ "refs/heads/" ++ show branchname
+fullname = Git.Ref $ "refs/heads/" ++ Git.Types.fromRef branchname
 
 originname :: Git.Ref
-originname = Git.Ref $ "refs/remotes/origin/" ++ show branchname
+originname = Git.Ref $ "refs/remotes/origin/" ++ Git.Types.fromRef branchname
 
 hasOrigin :: Backup Bool
 hasOrigin = inRepo $ Git.Ref.exists originname
@@ -580,8 +582,8 @@ runDeferred = go =<< getState deferredBackups
 		-- more actions to be deferred; run them too.
 		runDeferred
 
-backupName :: String -> IO ()
-backupName name = do
+backupOwner :: [GithubUserRepo] -> Owner -> IO ()
+backupOwner exclude (Owner name) = do
 	auth <- getAuth
 	l <- sequence
 	 	[ Github.userRepos' auth name Github.All
@@ -596,16 +598,7 @@ backupName name = do
 			else error $ "No GitHub repositories found for " ++ name
 	-- Clone any missing repos, and get a BackupState for each repo
 	-- that is to be backed up.
-	states <- forM nameurls $ \(dir, url) -> do
-		unlessM (doesDirectoryExist dir) $ do
-			putStrLn $ "New repository: " ++ dir
-			ok <- boolSystem "git"
-				[ Param "clone"
-				, Param url
-				, Param dir
-				]
-			unless ok $ error "clone failed"
-		genBackupState =<< Git.Construct.fromPath dir
+	states <- catMaybes <$> forM nameurls prepare
 	-- First pass only retries things that failed before, so the
 	-- retried actions will run in each repo before too much API is
 	-- used up.
@@ -613,15 +606,62 @@ backupName name = do
 	states'' <- forM states' (execStateT . runBackup $ mainBackup)
 	forM states'' (evalStateT . runBackup $ runDeferred >> save)
 		>>= showFailures . concat
+  where
+	excludeurls = map repoUrl exclude
+	prepare (dir, url)
+		| url `elem` excludeurls = return Nothing
+		| otherwise = do
+			print url
+			unlessM (doesDirectoryExist dir) $ do
+				putStrLn $ "New repository: " ++ dir
+				ok <- boolSystem "git"
+					[ Param "clone"
+					, Param url
+					, Param dir
+					]
+				unless ok $ error "clone failed"
+			Just <$> (genBackupState =<< Git.Construct.fromPath dir)
 
-usage :: String
-usage = "usage: github-backup [username|organization]"
+data Options = Options
+	{ includeOwner :: [Owner]
+	, excludeRepo :: [GithubUserRepo]
+	}
+	deriving (Show)
+
+data Owner = Owner String
+	deriving (Show)
+
+options :: Parser Options
+options = Options <$> many owneropt <*> many excludeopt
+  where
+	owneropt = (argument (Just . Owner))
+		( metavar "USERNAME|ORGANIZATION"
+		<> help "Back up repositories owned by this entity."
+		)
+	excludeopt = parseUserRepo <$> (strOption
+		( long "exclude"
+		<> metavar "USERNAME/REPOSITORY"
+		<> help "Skip backing up a repository."
+		))
+
+parseUserRepo :: String -> GithubUserRepo
+parseUserRepo s =
+	let (user, repo) = separate (== '/') s
+	in GithubUserRepo user repo
 
 main :: IO ()
-main = getArgs >>= go
+main = execParser opts >>= go
   where
-	go (('-':_):_) = error usage
-	go [] = backupRepo =<< Git.Construct.fromCwd
-	go (name:[]) = backupName name
-	go _= error usage
-
+	opts = info (helper <*> options)
+		( fullDesc
+		<> progDesc desc
+		<> header "github-backup - backs up data from GitHub"
+		)
+	desc = unlines
+		[ "Backs up all forks, issues, etc of a GitHub repository."
+		, "Run without any parameters inside a clone of a repository to back it up."
+		, "Or, specify whose repositories to back up."
+		]
+	go (Options owner exclude)
+		| null owner = backupRepo =<< Git.Construct.fromCwd
+		| otherwise = mapM_ (backupOwner exclude) owner
